/*! ********************************************
 * SynthesizerPCM.java
 * Copyright 2019 masahiro takamoto (MotChan)
 * Distributed under the MIT license. 
 * see -> sora/license.txt
 ******************************************** */
package sora.graybox.synth;

import sora.graybox.MidiEvent;
// PCM WaveTable Synthesis
public final class SynthesizerPCM extends MultiTimberSynthBase
{
	static final int PG_PHASE_FP = 18;  // 18bit fixed-point  (precision 1 / 262144)

private class VoiceUnitPCM extends VoiceUnitBase
{
// Tables
	// Log2 Volume Table
	private float[] _LOG2V_TABLE = null;
	// Pow2 Table
	private final int _POW2T_OCTV_BITS  = SynthData.POW2T_OCTV_BITS;
	private final int _POW2T_OCTV_MASK  = SynthData.POW2T_OCTV_MASK;
	private int[] _POW2_TABLE      = null;
	private int[] _POW2T_RS_TABLE  = null;
	private int[] _POW2T_KEY_TABLE = null;
	// EG
	private final int _EG_PRAM_OCTV_LS = SynthData.POW2T_OCTV_BITS - 3;
	private final int _EG_CNT_FP       = SynthData.EG_CNT_FP;
	private final int _EG_MAX_CNT      = SynthData.EG_MAX_CNT;
	private final int _EG_CNT_ENV_ATK  = SynthData.EG_CNT_ENV_ATK;
	private final int _EG_CNT_ENV_BTM  = SynthData.EG_CNT_ENV_BTM;
	private final int _EG_SUSTAIN_BTM  = SynthData.EG_SUSTAIN_BTM;
	private int[] _EG_TOTAL_LEVEL_TABLE = null;
	private int[] _EG_LEVEL_TABLE_UP    = null;
	private int[] _EG_LEVEL_TABLE_DW    = null;
	private int[] _EG_KEY_TABLE         = null;
	private int[] _EG_POW2T_RS_TABLE    = null;
	private int[] _EG_CNT_GATE_TABLE    = null;
	// PG
	private final int _PG_PHASE_FP      = SynthesizerPCM.PG_PHASE_FP;
	private final int _PG_PHASE_FP_MASK = (1 << _PG_PHASE_FP) - 1;
	// LFO
	private final int _LFO_TIMER_INIT_VALUE = SynthData.LFO_TIMER_INIT_VALUE;
	private final int _LFO_WAVE_TABLE_MASK  = SynthData.LFO_WAVE_TABLE_SIZE - 1;
	private int[] _LFO_WAVE_TABLE = null;
// Synthesizer Data
	VoicePart[] VOICE_PARTS = null;
	SampleData[] SAMPLE_DATAS = null;
	TonePresetPCM[] TONE_PRESETS = null;
	SamplerPreset[] SAMPLER_PRESETS = null;
	int pgSamplePow2tIdx = 0;  // PG PhaseStep (Pow2Table Index Offset) - SamplingRate
	int pgKeyPow2tIdx    = 0;  // PG PhaseStep (Pow2Table Index Offset) - ReferenceKey (NoteNumber 0)
	int egPow2tIdx       = 0;  // EG PhaseStep (Pow2Table Index Offset)
	int egPhaseStepSndOff= 0;
	int lfoTimerCoef     = 0;  // Coefficient Value - LFO Phase Update Timer
	// Pitch
	private int _ptBend     = 0;   // pitchBend
	private int _ptBendSens = 0;   // pitchBend Sensitivity
	// LFO
	private int _lfoPhase     = 0;  // LFO Phase
	private int _lfoTimer     = 0;  // LFO Phase Update Timer
	private int _lfoTimerStep = 0;  // Update Timer Step
	private int _lfoMod       = 0;  // Modulation Depth
	private int _lfoModSens   = 0;  // Modulation Sensitivity
	private int _lfoOutPM     = 0;  // output - phase modulation
	private int _lfoOutAM     = 0;  // output - amplitude modulation
	// Wave
	private int[] _waveTable = null;
	// EG
	/*
	- Envelop Generator State -
	[ 5674 3210 ] 8Bits Flag Status (0 = deActive)
	[ ---- RSDA ] "-" = Reserved Bits
		A = Attack  stage
		D = Decay   stage
		S = Sustain stage
		R = Release stage
	*/
	private int _egState      = 0;
	private int[] _egLevelTable = null;
	private int _egCount      = 0;
	private int _egCountStep  = 0;
	//private int _egCountStepAR= 0;
	private int _egCountStepDR= 0;
	private int _egCountStepRR= 0;
	private int _egCntEnvSus  = 0;
	private int _egCntEnvBtm  = 0;
	private int _egGate       = 0;
	private int _egSL         = 0;
	private int _egTL         = 0;
	private int _egTLv        = 0;
	// PG
	private int _pgKey       = 0;
	private int _pgBend      = 0;
	private int _pgPhase     = 0;
	private int _pgPhaseAFP  = 0;  // PG Phase (After the Fixed Point)
	private int _pgPhaseStep = 0;
// Sampler
	private int _plbkE      = 0;  // PlayBack End Frame   [plbkE <= channelPerSamples]
	private int _loopE      = 0;  // Loop End Frame       [loopE <= plbkE]
	private int _loopF      = 0;  // Loop Frames          [val = loopE - lpStart]
	private int _loopFk     = 0;  // Loop Frames          * after -> keyOff
	
	@Override
	void keyOff()
	{
		// KeyOn --->
		if ((state & 0x0010) != 0)
		{
			// ATTACK ->
			if (_egState == 1)
			{
				_egLevelTable = _EG_LEVEL_TABLE_DW;
				if (_egCount >= _EG_MAX_CNT)_egCount = 0;
				else _egCount = _EG_LEVEL_TABLE_UP[_egCount >> _EG_CNT_FP] << _EG_CNT_FP - 1;  // -1 = 8192 to 4096
				if (_egCount > _egCntEnvBtm){ killProcess(); return; }
			}
			// DECAY,SUSTAIN ->
			_egGate = _egCntEnvBtm;
			_egCountStep = _egCountStepRR;
			_egState = 8;  // Release
			state = 0x0021;
			// sampler ---
			_loopE = _plbkE;
			_loopF = _loopFk;
		}
	}
	@Override
	void soundOff()
	{
		// KeyOn || KeyOff --->
		if ((state & 0x0030)!= 0)
		{
			// ATTACK ->
			if (_egState == 1)
			{
				_egLevelTable = _EG_LEVEL_TABLE_DW;
				if (_egCount >= _EG_MAX_CNT)_egCount = 0;
				else _egCount = _EG_LEVEL_TABLE_UP[_egCount >> _EG_CNT_FP] << _EG_CNT_FP - 1;
				if (_egCount > _egCntEnvBtm){ killProcess(); return; }
			}
			// DECAY,SUSTAIN ->
			_egGate = _egCntEnvBtm;
			_egCountStep = egPhaseStepSndOff;
			_egState = 0;  // SoundOff, EG OFF
			partNumber = -1; // SoundOff
			state = 0x0041;
			// sampler ---
			_loopE = _plbkE;
			_loopF = _loopFk;
		}
	}
	private int _shiftEGState()
	{
		switch (_egState)
		{
			// ATTACK ->
			case 1:
				_egLevelTable = _EG_LEVEL_TABLE_DW;
				_egCount = 0;
				// -> DECAY
				if (_egSL > 0)
				{
					_egGate = _egCntEnvSus;
					_egCountStep = _egCountStepDR;
					return _egState = 2;
				}
				// -> SUSTAIN
				_egGate = 0x7fffffff;
				_egCountStep = 0;
				return _egState = 4;
			// DECAY ->
			case 2:
				// -> SUSTAIN
				if ((_egTL + _egSL) < _EG_SUSTAIN_BTM)
				{
					_egLevelTable = _EG_LEVEL_TABLE_DW;
					_egGate = 0x7fffffff;
					_egCountStep = 0;
					_egCount = _egCntEnvSus;
					return _egState = 4;
				}
			// DECAY,RELEASE,SoundOFF -> EG OFF
			default: return _egState = 0;
		}
	}
	@Override
	void keyOn(int ch, int note, int velo)
	{
	// KeyOn Status ---
		state = 0x0011;
		partNumber = ch;
		noteNumber = note;
	// Part, Preset ---
		VoicePart _part = VOICE_PARTS[partNumber]; 
		TonePresetPCM _preset = TONE_PRESETS[_part.programChange];
		SampleData _sample = SAMPLE_DATAS[_preset.smplNum];
		if (_sample.sampleData == null){ killProcess(); return; }
		_waveTable = _sample.sampleData;
		_plbkE = _sample.plbkE;
		_loopE = _sample.loopE;
		_loopF = _sample.loopF;
		_loopFk = _sample.loopFk;
	// LFO ---
		_lfoModSens = _preset.lms;
		_lfoMod = _part.ccValues[1] * _lfoModSens;
		_lfoTimerStep = _preset.lfoFreq * lfoTimerCoef;
		_lfoTimer = _LFO_TIMER_INIT_VALUE;
		_lfoPhase = _preset.strc + 1 & 1;
		_lfoOutPM = 0;
		_lfoOutAM = 0;
	// volume ---
		_vPtVol = _mainVolTable[_part.ccValues[7]];
		_vPtExp = _exprVolTable[_part.ccValues[11]];
		_vVel = _veloVolTable[velo];
		_vVol = _vVel * _masterVolume;
		_vOut = _vPtVol * _vPtExp * _vVol;
		_prPan = 64;
		_vPanL = _panTable[      _prPan + _part.ccValues[10]];
		_vPanR = _panTable[256 - _prPan - _part.ccValues[10]];
		_vOutL = _vOut * _vPanL;
		_vOutR = _vOut * _vPanR;
	// PG ---
		_ptBendSens = _POW2T_KEY_TABLE[_part.bendSens];
		_ptBend = _part.pitchBend - 8192;
		_pgBend = _ptBend * _ptBendSens >> 13;
		_pgKey = _POW2T_KEY_TABLE[noteNumber] + (_preset.tune - 8192) + _sample.pgKey + pgKeyPow2tIdx;
		int v = _pgKey + _pgBend;// & 0x7fffffff;
		_pgPhaseStep = _POW2_TABLE[v & _POW2T_OCTV_MASK] >> _POW2T_RS_TABLE[v >> _POW2T_OCTV_BITS];
		_pgPhase = _sample.plbkS;
		_pgPhaseAFP = 0;
	// EG ---
		_egTL = _preset.tl;
		_egTLv = _EG_TOTAL_LEVEL_TABLE[_egTL];
		_egCntEnvBtm = _EG_CNT_GATE_TABLE[128 - _egTL];
		_egSL = _preset.sl;
		_egCntEnvSus = _EG_CNT_GATE_TABLE[_egSL];
		if (_egCntEnvSus > _egCntEnvBtm)_egCntEnvSus = _egCntEnvBtm;
		int _keyFlw = _EG_KEY_TABLE[127 + noteNumber - _preset.kfStart] << _preset.kf;
		_keyFlw += egPow2tIdx;
		v = (_preset.dr << _EG_PRAM_OCTV_LS) + _keyFlw;
		_egCountStepDR = _POW2_TABLE[v & _POW2T_OCTV_MASK] >> _EG_POW2T_RS_TABLE[v >> _POW2T_OCTV_BITS];
		v = (_preset.rr << _EG_PRAM_OCTV_LS) + _keyFlw;
		_egCountStepRR = _POW2_TABLE[v & _POW2T_OCTV_MASK] >> _EG_POW2T_RS_TABLE[v >> _POW2T_OCTV_BITS];
	// EG State ---
		// -> ATTACK
		if (_preset.ar < 127) {
			_egCount = _EG_CNT_ENV_ATK;
			_egLevelTable = _EG_LEVEL_TABLE_UP;
			_egGate = _EG_MAX_CNT;
			v = (_preset.ar << _EG_PRAM_OCTV_LS) + _keyFlw;
			_egCountStep = _POW2_TABLE[v & _POW2T_OCTV_MASK] >> _EG_POW2T_RS_TABLE[v >> _POW2T_OCTV_BITS];
			_egState = 1;
		} else {
			_egLevelTable = _EG_LEVEL_TABLE_DW;
			_egCount = 0;
		// -> DECAY
			if (_egSL > 0) {
				_egGate = _egCntEnvSus;
				_egCountStep = _egCountStepDR;
				_egState = 2;
		// -> SUSTAIN
			} else {
				_egGate = 0x7fffffff;  // dummy
				_egCountStep = 0;  // dummy
				_egState = 4; }
		}
	// process ---
		v = (_preset.strc & 3) << 1;
		if (_sample.channelCount == 2)v|=1;
		process = _PROC_PTRS[v];
	}
	void keyOnSampler(int ch, int note, int velo)
	{
	// KeyOn Status ---
		state = 0x0011;
		partNumber = ch;
		noteNumber = note;
	// tone ---
		VoicePart _part = VOICE_PARTS[partNumber];
		SamplerPreset _preset = SAMPLER_PRESETS[noteNumber];
		SampleData _sample = SAMPLE_DATAS[_preset.smplNum];
		if (_sample.sampleData == null){ killProcess(); return; }
		_waveTable = _sample.sampleData;
		_plbkE = _sample.plbkE;
		_loopE = _sample.loopE;
		_loopF = _sample.loopF;
		_loopFk = _sample.loopFk;
	// LFO ---
		_lfoOutPM = 0;
		_lfoOutAM = 0;
	// volume ---
		_vPtVol = _mainVolTable[_part.ccValues[7]];
		_vPtExp = _exprVolTable[_part.ccValues[11]];
		_vVel = _veloVolTable[velo];
		_vVol = _vVel * _masterVolume;
		_vOut = _vPtVol * _vPtExp * _vVol;
		_prPan = _preset.pan;
		_vPanL = _panTable[      _prPan + _part.ccValues[10]];
		_vPanR = _panTable[256 - _prPan - _part.ccValues[10]];
		_vOutL = _vOut * _vPanL;
		_vOutR = _vOut * _vPanR;
	// PG ---
		_ptBendSens = _POW2T_KEY_TABLE[_part.bendSens];
		_ptBend = _part.pitchBend - 8192;
		_pgBend = _ptBend * _ptBendSens >> 13;
		_pgKey = (_preset.octv - 8 << 13) + (_preset.tune - 8192) + _sample.pgSample + pgSamplePow2tIdx;
		int v = _pgKey + _pgBend;
		_pgPhaseStep = _POW2_TABLE[v & _POW2T_OCTV_MASK] >> _POW2T_RS_TABLE[v >> _POW2T_OCTV_BITS];
		_pgPhase = _sample.plbkS;
		_pgPhaseAFP = 0;
	// EG ---
		_egTL = _preset.tl;
		_egTLv = _EG_TOTAL_LEVEL_TABLE[_egTL];
		_egCntEnvBtm = _EG_CNT_GATE_TABLE[128 - _egTL];
		_egSL = _preset.sl;
		_egCntEnvSus = _EG_CNT_GATE_TABLE[_egSL];
		if (_egCntEnvSus > _egCntEnvBtm)_egCntEnvSus = _egCntEnvBtm;
		v = (_preset.dr << _EG_PRAM_OCTV_LS) + egPow2tIdx;
		_egCountStepDR = _POW2_TABLE[v & _POW2T_OCTV_MASK] >> _EG_POW2T_RS_TABLE[v >> _POW2T_OCTV_BITS];
		v = (_preset.rr << _EG_PRAM_OCTV_LS) + egPow2tIdx;
		_egCountStepRR = _POW2_TABLE[v & _POW2T_OCTV_MASK] >> _EG_POW2T_RS_TABLE[v >> _POW2T_OCTV_BITS];
	// EG State ---
		// -> ATTACK
		if (_preset.ar < 127) {
			_egCount = _EG_CNT_ENV_ATK;
			_egLevelTable = _EG_LEVEL_TABLE_UP;
			_egGate = _EG_MAX_CNT;
			v = (_preset.ar << _EG_PRAM_OCTV_LS) + egPow2tIdx;
			_egCountStep = _POW2_TABLE[v & _POW2T_OCTV_MASK] >> _EG_POW2T_RS_TABLE[v >> _POW2T_OCTV_BITS];
			_egState = 1;
		} else {
			_egLevelTable = _EG_LEVEL_TABLE_DW;
			_egCount = 0;
		// -> DECAY
			if (_egSL > 0) {
				_egGate = _egCntEnvSus;
				_egCountStep = _egCountStepDR;
				_egState = 2;
		// -> SUSTAIN
			} else {
				_egGate = 0x7fffffff;  // dummy
				_egCountStep = 0;  // dummy
				_egState = 4; }
		}
	// process ---
		process = (_sample.channelCount == 2)? this::_procSmpl2ch: this::_procSmpl1ch;
	}
	@Override
	void updPartBendSens(int val)
	{
		_ptBendSens = _POW2T_KEY_TABLE[val];
		_pgBend = _ptBend * _ptBendSens >> 13;
		int v = _pgKey + _pgBend + _lfoOutPM;
		_pgPhaseStep = _POW2_TABLE[v & _POW2T_OCTV_MASK] >> _POW2T_RS_TABLE[v >> _POW2T_OCTV_BITS];
	}
	@Override
	void updPartPitchBend(int val)
	{
		_ptBend = val - 8192;
		_pgBend = _ptBend * _ptBendSens >> 13;
		int v = _pgKey + _pgBend + _lfoOutPM;
		_pgPhaseStep = _POW2_TABLE[v & _POW2T_OCTV_MASK] >> _POW2T_RS_TABLE[v >> _POW2T_OCTV_BITS];
	}
	@Override
	void updPartModulation(int val)
	{
		_lfoMod = val * _lfoModSens;
	}
	void updateLfoFrequency(int val)
	{
		_lfoTimerStep = val * lfoTimerCoef;
	}
	void setup(SynthData synthData)
	{
		// volume Table
		_mainVolTable = synthData.getVOLUME_TABLE_0dB();
		_exprVolTable = synthData.getVOLUME_TABLE_0dB();
		_veloVolTable = synthData.getVOLUME_TABLE_0dB();
		_panTable     = synthData.getPAN_TABLE_SIN();
		// Log2 Volume Table
		_LOG2V_TABLE = synthData.getLOG2V_TABLE();
		// Pow2 Table
		_POW2_TABLE      = synthData.getPOW2_TABLE();
		_POW2T_RS_TABLE  = synthData.getPOW2T_RS_TABLE();
		_POW2T_KEY_TABLE = synthData.getPOW2T_KEY_TABLE();
		// EG
		_EG_TOTAL_LEVEL_TABLE = synthData.getEG_TOTAL_LEVEL_TABLE();
		_EG_LEVEL_TABLE_UP    = synthData.getEG_LEVEL_TABLE_UP();
		_EG_LEVEL_TABLE_DW    = synthData.getEG_LEVEL_TABLE_DW();
		_EG_KEY_TABLE         = synthData.getEG_KEY_TABLE();
		_EG_POW2T_RS_TABLE    = synthData.getEG_POW2T_RS_TABLE();
		_EG_CNT_GATE_TABLE    = synthData.getEG_CNT_GATE_TABLE();
		// etc
		_LFO_WAVE_TABLE = synthData.getLFO_WAVE_TABLES()[0];
	}
// Constructor
	VoiceUnitPCM()
	{
		super();
		_PROC_PTRS[0] = this::_proc0;
		_PROC_PTRS[1] = this::_proc1;
		_PROC_PTRS[2] = this::_proc2;
		_PROC_PTRS[3] = this::_proc3;
		_PROC_PTRS[4] = this::_proc4;
		_PROC_PTRS[5] = this::_proc5;
		_PROC_PTRS[6] = this::_proc0;
		_PROC_PTRS[7] = this::_proc1;
	}
// process ---
	// pointer array (process Function)
	private final Function_PRC[] _PROC_PTRS = new Function_PRC[8];
	// Inst 1ch
	private void _proc0(float[] outL, float[] outR, int end)
	{
		int[]  _et = _egLevelTable;
		final float[] _log2vT = _LOG2V_TABLE;
		final int[] _wt = _waveTable;
		int i = startFrame;
		for ( ; i < end; i++)
		{
		// eg update --->
			if (_egCount >= _egGate)
			{
				if (_shiftEGState() != 0)_et = _egLevelTable;
				else { killProcess(); return; }
			}
			int v = _et[_egCount >> _EG_CNT_FP] + _egTLv;	
		// pg update --->
			if (_pgPhase >= _loopE)
			{
				if (_loopF > 0) do{ _pgPhase -= _loopF; } while (_pgPhase >= _loopE);
				else { killProcess(); return; }
			}
			v = _wt[_pgPhase] + v;
			float n = _log2vT[v];
		// out --->
			outL[i] += n * _vOutL;
			outR[i] += n * _vOutR;
		// counter --->
			_egCount += _egCountStep;
			_pgPhaseAFP += _pgPhaseStep;
			_pgPhase += _pgPhaseAFP >> _PG_PHASE_FP;
			_pgPhaseAFP &= _PG_PHASE_FP_MASK;
		}
		startFrame = i;
	}
	// Inst 2ch
	private void _proc1(float[] outL, float[] outR, int end)
	{
		int[]  _et = _egLevelTable;
		final float[] _log2vT = _LOG2V_TABLE;
		final int[] _wt = _waveTable;
		int i = startFrame;
		for ( ; i < end; i++)
		{
		// eg update --->
			if (_egCount >= _egGate)
			{
				if (_shiftEGState() != 0)_et = _egLevelTable;
				else { killProcess(); return; }
			}
			int v = _et[_egCount >> _EG_CNT_FP] + _egTLv;	
		// pg update --->
			if (_pgPhase >= _loopE)
			{
				if (_loopF > 0) do{ _pgPhase -= _loopF; } while (_pgPhase >= _loopE);
				else { killProcess(); return; }
			}
		// out --->
			int p = _pgPhase << 1;
			outL[i] += _log2vT[_wt[p    ] + v] * _vOutL;
			outR[i] += _log2vT[_wt[p + 1] + v] * _vOutR;
		// counter --->
			_egCount += _egCountStep;
			_pgPhaseAFP += _pgPhaseStep;
			_pgPhase += _pgPhaseAFP >> _PG_PHASE_FP;
			_pgPhaseAFP &= _PG_PHASE_FP_MASK;
		}
		startFrame = i;
	}
	// Inst 1ch + LFO PM
	private void _proc2(float[] outL, float[] outR, int end)
	{
		int[]  _et = _egLevelTable;
		final float[] _log2vT = _LOG2V_TABLE;
		final int[] _wt = _waveTable;
		int i = startFrame;
		for ( ; i < end; i++)
		{
		// lfo Update --->
			if (_lfoTimer <= 0)
			{
				_lfoOutPM = (_LFO_WAVE_TABLE[_lfoPhase] * _lfoMod) >> 11;  // 7(mod) + 4(sens)
				int l = _pgKey + _pgBend + _lfoOutPM;
				_pgPhaseStep = _POW2_TABLE[l & _POW2T_OCTV_MASK] >> _POW2T_RS_TABLE[l >> _POW2T_OCTV_BITS];
				_lfoPhase = _lfoPhase + 2 & _LFO_WAVE_TABLE_MASK;
				_lfoTimer += _LFO_TIMER_INIT_VALUE;
			}
		// eg update --->
			if (_egCount >= _egGate)
			{
				if (_shiftEGState() != 0)_et = _egLevelTable;
				else { killProcess(); return; }
			}
			int v = _et[_egCount >> _EG_CNT_FP] + _egTLv;	
		// pg update --->
			if (_pgPhase >= _loopE)
			{
				if (_loopF > 0) do{ _pgPhase -= _loopF; } while (_pgPhase >= _loopE);
				else { killProcess(); return; }
			}
			v = _wt[_pgPhase] + v;
			float n = _log2vT[v];
		// out --->
			outL[i] += n * _vOutL;
			outR[i] += n * _vOutR;
		// counter --->
			_lfoTimer -= _lfoTimerStep;
			_egCount += _egCountStep;
			_pgPhaseAFP += _pgPhaseStep;
			_pgPhase += _pgPhaseAFP >> _PG_PHASE_FP;
			_pgPhaseAFP &= _PG_PHASE_FP_MASK;
		}
		startFrame = i;
	}
	// Inst 2ch + LFO PM
	private void _proc3(float[] outL, float[] outR, int end)
	{
		int[]  _et = _egLevelTable;
		final float[] _log2vT = _LOG2V_TABLE;
		final int[] _wt = _waveTable;
		int i = startFrame;
		for ( ; i < end; i++)
		{
		// lfo Update --->
			if (_lfoTimer <= 0)
			{
				_lfoOutPM = (_LFO_WAVE_TABLE[_lfoPhase] * _lfoMod) >> 11;  // 7(mod) + 4(sens)
				int l = _pgKey + _pgBend + _lfoOutPM;
				_pgPhaseStep = _POW2_TABLE[l & _POW2T_OCTV_MASK] >> _POW2T_RS_TABLE[l >> _POW2T_OCTV_BITS];
				_lfoPhase = _lfoPhase + 2 & _LFO_WAVE_TABLE_MASK;
				_lfoTimer += _LFO_TIMER_INIT_VALUE;
			}
		// eg update --->
			if (_egCount >= _egGate)
			{
				if (_shiftEGState() != 0)_et = _egLevelTable;
				else { killProcess(); return; }
			}
			int v = _et[_egCount >> _EG_CNT_FP] + _egTLv;	
		// pg update --->
			if (_pgPhase >= _loopE)
			{
				if (_loopF > 0) do{ _pgPhase -= _loopF; } while (_pgPhase >= _loopE);
				else { killProcess(); return; }
			}
		// out --->
			int p = _pgPhase << 1;
			outL[i] += _log2vT[_wt[p    ] + v] * _vOutL;
			outR[i] += _log2vT[_wt[p + 1] + v] * _vOutR;
		// counter --->
			_lfoTimer -= _lfoTimerStep;
			_egCount += _egCountStep;
			_pgPhaseAFP += _pgPhaseStep;
			_pgPhase += _pgPhaseAFP >> _PG_PHASE_FP;
			_pgPhaseAFP &= _PG_PHASE_FP_MASK;
		}
		startFrame = i;
	}
	// Inst 1ch + LFO AM
	private void _proc4(float[] outL, float[] outR, int end)
	{
		int[]  _et = _egLevelTable;
		final float[] _log2vT = _LOG2V_TABLE;
		final int[] _wt = _waveTable;
		int i = startFrame;
		for ( ; i < end; i++)
		{
		// lfo Update --->
			if (_lfoTimer <= 0)
			{
				int l = (_LFO_WAVE_TABLE[_lfoPhase] * _lfoMod) >> 11;
				_lfoOutAM = l & 0x7ffffffe;
				_lfoPhase = _lfoPhase + 2 & _LFO_WAVE_TABLE_MASK;
				_lfoTimer += _LFO_TIMER_INIT_VALUE;
			}
		// eg update --->
			if (_egCount >= _egGate)
			{
				if (_shiftEGState() != 0)_et = _egLevelTable;
				else { killProcess(); return; }
			}
			int v = _et[_egCount >> _EG_CNT_FP] + _egTLv;	
		// pg update --->
			if (_pgPhase >= _loopE)
			{
				if (_loopF > 0) do{ _pgPhase -= _loopF; } while (_pgPhase >= _loopE);
				else { killProcess(); return; }
			}
			v = _wt[_pgPhase] + v + _lfoOutAM;
			float n = _log2vT[v];
		// out --->
			outL[i] += n * _vOutL;
			outR[i] += n * _vOutR;
		// counter --->
			_lfoTimer -= _lfoTimerStep;
			_egCount += _egCountStep;
			_pgPhaseAFP += _pgPhaseStep;
			_pgPhase += _pgPhaseAFP >> _PG_PHASE_FP;
			_pgPhaseAFP &= _PG_PHASE_FP_MASK;
		}
		startFrame = i;
	}
	// Inst 2ch + LFO AM
	private void _proc5(float[] outL, float[] outR, int end)
	{
		int[]  _et = _egLevelTable;
		final float[] _log2vT = _LOG2V_TABLE;
		final int[] _wt = _waveTable;
		int i = startFrame;
		for ( ; i < end; i++)
		{
		// lfo Update --->
			if (_lfoTimer <= 0)
			{
				int l = (_LFO_WAVE_TABLE[_lfoPhase | 1] * _lfoMod) >> 11;
				_lfoOutAM = l & 0x7ffffffe;
				_lfoPhase = _lfoPhase + 2 & _LFO_WAVE_TABLE_MASK;
				_lfoTimer += _LFO_TIMER_INIT_VALUE;
			}
		// eg update --->
			if (_egCount >= _egGate)
			{
				if (_shiftEGState() != 0)_et = _egLevelTable;
				else { killProcess(); return; }
			}
			int v = _et[_egCount >> _EG_CNT_FP] + _egTLv + _lfoOutAM;	
		// pg update --->
			if (_pgPhase >= _loopE)
			{
				if (_loopF > 0) do{ _pgPhase -= _loopF; } while (_pgPhase >= _loopE);
				else { killProcess(); return; }
			}
		// out --->
			int p = _pgPhase << 1;
			outL[i] += _log2vT[_wt[p    ] + v] * _vOutL;
			outR[i] += _log2vT[_wt[p + 1] + v] * _vOutR;
		// counter --->
			_lfoTimer -= _lfoTimerStep;
			_egCount += _egCountStep;
			_pgPhaseAFP += _pgPhaseStep;
			_pgPhase += _pgPhaseAFP >> _PG_PHASE_FP;
			_pgPhaseAFP &= _PG_PHASE_FP_MASK;
		}
		startFrame = i;
	}
	// Sampler 1ch
	private void _procSmpl1ch(float[] outL, float[] outR, int end)
	{
		int[]  _et = _egLevelTable;
		final float[] _log2vT = _LOG2V_TABLE;
		final int[] _wt = _waveTable;
		int i = startFrame;
		for ( ; i < end; i++)
		{
		// eg update --->
			if (_egCount >= _egGate)
			{
				if (_shiftEGState() != 0)_et = _egLevelTable;
				else { killProcess(); return; }
			}
			int v = _et[_egCount >> _EG_CNT_FP] + _egTLv;	
		// pg update --->
			if (_pgPhase >= _loopE)
			{
				if (_loopF > 0) do{ _pgPhase -= _loopF; } while (_pgPhase >= _loopE);
				else { killProcess(); return; }
			}
			v = _wt[_pgPhase] + v;
			float n = _log2vT[v];
		// out --->
			outL[i] += n * _vOutL;
			outR[i] += n * _vOutR;
		// counter --->
			_egCount += _egCountStep;
			_pgPhaseAFP += _pgPhaseStep;
			_pgPhase += _pgPhaseAFP >> _PG_PHASE_FP;
			_pgPhaseAFP &= _PG_PHASE_FP_MASK;
		}
		startFrame = i;
	}
	// Sampler 2ch
	private void _procSmpl2ch(float[] outL, float[] outR, int end)
	{
		int[]  _et = _egLevelTable;
		final float[] _log2vT = _LOG2V_TABLE;
		final int[] _wt = _waveTable;
		int i = startFrame;
		for ( ; i < end; i++)
		{
		// eg update --->
			if (_egCount >= _egGate)
			{
				if (_shiftEGState() != 0)_et = _egLevelTable;
				else { killProcess(); return; }
			}
			int v = _et[_egCount >> _EG_CNT_FP] + _egTLv;	
		// pg update --->
			if (_pgPhase >= _loopE)
			{
				if (_loopF > 0) do{ _pgPhase -= _loopF; } while (_pgPhase >= _loopE);
				else { killProcess(); return; }
			}
		// out --->
			int p = _pgPhase << 1;
			outL[i] += _log2vT[_wt[p    ] + v] * _vOutL;
			outR[i] += _log2vT[_wt[p + 1] + v] * _vOutR;
		// counter --->
			_egCount += _egCountStep;
			_pgPhaseAFP += _pgPhaseStep;
			_pgPhase += _pgPhaseAFP >> _PG_PHASE_FP;
			_pgPhaseAFP &= _PG_PHASE_FP_MASK;
		}
		startFrame = i;
	}
}
// Factory -----
	@Override
	protected void _recreateVU(int maxPoly)
	{
		_voiceUnits = new VoiceUnitBase[maxPoly];
		_voiceUnitExt = new VoiceUnitPCM[maxPoly];
		// Reset VoiceUnit
		for (int i = 0; i < maxPoly; i++)
		{
			_voiceUnitExt[i] = new VoiceUnitPCM();
			VoiceUnitPCM _vu = _voiceUnitExt[i];
			_voiceUnits[i] = _vu;
			_vu.setup(_synthData);
			_vu.killProcess();
			_vu.VOICE_PARTS = voiceParts;
			_vu.SAMPLE_DATAS = sampleDatas;
			_vu.TONE_PRESETS = tonePresets;
			_vu.SAMPLER_PRESETS = samplerPresets;
			_vu.setMasterVolume(_masterVolume);
		}
		// set Property
		_maxPolyphony = maxPoly;
	}
	@Override
	protected void _resetKeyParams()
	{
		final double _sampleRateD = (double)_sampleRate;
		// LFO
		double _lfoWaveSize = SynthData.LFO_WAVE_TABLE_SIZE / 2.0;
			// Max Frame Rate - LFO Phase Step
		double _lfoMaxFrameRate = _sampleRateD / SynthData.LFO_MAX_FREQ / _lfoWaveSize;
		if (_lfoMaxFrameRate < 1.0)_lfoMaxFrameRate = 1.0;  // Max 1frame per 1Phase
		int _lfoTimerCoef =  (int)(SynthData.LFO_TIMER_INIT_VALUE / 16383.0 / _lfoMaxFrameRate + 0.5);
		// EG
		double _rate0Frames = SynthData.EG_REFERENCE_TIME * SynthData.LOG2VT_DYN_BITS * _sampleRateD;  // to -96dB Frames
		int _egPow2tIdx = SynthData.toPow2tIndexValue(SynthData.EG_MAX_CNT / _rate0Frames);
		double _sndOffFrames = SynthData.EG_SOUND_OFF_TIME / SynthData.LOG2VT_DYN_BITS * _sampleRateD;  // to -96dB Frames
		int _egSndOff = (int)(SynthData.EG_MAX_CNT / _sndOffFrames + 0.5);
		// PG PCM
		int _smpl = SynthData.toPow2tIndexValue(_sampleRateD);
		int _pgSamplePow2tIdx = SynthData.toPow2tIndexValue((double)(1 << PG_PHASE_FP)) - _smpl;
		// PG WaveTable
		int _pgKeyPow2tIdx = _pgSamplePow2tIdx + SynthData.toPow2tIndexValue(_referencePitch * Math.pow(2.0, -69.0 / 12.0)); // note 0 Pitch
		for (int i = 0; i < _maxPolyphony; i++)
		{
			VoiceUnitPCM _vu = _voiceUnitExt[i];
			_vu.killProcess();
			_vu.pgSamplePow2tIdx = _pgSamplePow2tIdx;
			_vu.pgKeyPow2tIdx = _pgKeyPow2tIdx;
			_vu.egPow2tIdx = _egPow2tIdx;
			_vu.egPhaseStepSndOff = _egSndOff;
			_vu.lfoTimerCoef = _lfoTimerCoef;
		}
	}
// Synthesizer - Data -----
	private SampleData[] sampleDatas = null;
	public TonePresetPCM[] tonePresets = null;
	public SamplerPreset[] samplerPresets = null;
	private SynthData _synthData = null;
	public SampleData getSampleData(int id){ return sampleDatas[id]; }
	public void setSampleData(int id, SampleData sampleData, boolean deepCopy)
	{
		allSoundStop();
		sampleDatas[id].copy(sampleData, deepCopy);
	}
	public void loadPCMSamples(int id, float[] wave, int ch)
	{
		allSoundStop();
		sampleDatas[id].loadPCMSamples(wave, ch);
	}
	public void loadPCMData(int id, byte[] dat, int position, int sampleFrames, int ch, int bitRate, boolean sign, boolean bigEndian)
	{
		allSoundStop();
		sampleDatas[id].loadPCMData(dat, position, sampleFrames, ch, bitRate, sign, bigEndian);
	}
	public void clearSampleData(int id)
	{
		allSoundStop();
		sampleDatas[id].clear();
	}
	@Override
	public void factoryDataReset()
	{
		if (_voiceUnitExt != null) allSoundStop();
	// Sample
		if (sampleDatas == null)sampleDatas = new SampleData[128];
		for (int i = 0; i < 128; i++)
		{
			if (sampleDatas[i] == null) sampleDatas[i] = new SampleData();
			sampleDatas[i].clear();
		}
		if (tonePresets == null)tonePresets = new TonePresetPCM[128];
		if (samplerPresets == null)samplerPresets = new SamplerPreset[128];
		for (int i = 0; i < 128; i++)
		{
			if (tonePresets[i] == null)tonePresets[i] = new TonePresetPCM();
			if (samplerPresets[i] == null)samplerPresets[i] = new SamplerPreset();
			tonePresets[i].reset();
			samplerPresets[i].reset();
		}
	}
// Synthesizer - main -----
	private VoiceUnitPCM[] _voiceUnitExt = null;
// Constructor
	public SynthesizerPCM(){ this(32, 440.0f, 1.0f, true); }
	public SynthesizerPCM(int maxPoly, float refPitch, float masVol, boolean privateTableData)
	{
	// super ===
		super(maxPoly, refPitch, masVol);
	// SynthesizerPCM ===
		_modelID = 0x23;
		_maxPMN = 3;
		// create
		_synthData = (privateTableData)? new SynthData() : SynthData.getInstance();
		factoryDataReset();
		_recreateVU(_maxPolyphony);
		_resetKeyParams();
	}
// MidiEvent ---
	@Override
	protected void _noteOn(MidiEvent ev)
	{
		VoicePart _pt = voiceParts[ev.channel];
		final int _start = _nton;
		while ((_nton = _nton + 1 & _ntonM) != _start)
		{
			VoiceUnitPCM _vu = _voiceUnitExt[_nton];
			if (_vu.state == 0)
			{
				_vu.startFrame = ev.offsetFrames;
				if (_pt.mode == 3)_vu.keyOnSampler(ev.channel, ev.value0, ev.value1);
				else _vu.keyOn(ev.channel, ev.value0, ev.value1);
				return;
			}
		}
			
		while ((_nton = _nton + 1 & _ntonM) != _start)
		{
			VoiceUnitPCM _vu = _voiceUnitExt[_nton];
			// (KeyOff || SoundOff) -->
			if ((_vu.state & 0x60) != 0)
			{
				_vu.startFrame = ev.offsetFrames;
				if (_pt.mode == 3)_vu.keyOnSampler(ev.channel, ev.value0, ev.value1);
				else _vu.keyOn(ev.channel, ev.value0, ev.value1);
				return;
			}
		}
		
		while ((_nton = _nton + 1 & _ntonM) != _start)
		{
			VoiceUnitPCM _vu = _voiceUnitExt[_nton];
			// KeyOn -->
			if ((_vu.state & 0x10) != 0)
			{
				_vu.startFrame = ev.offsetFrames;
				if (_pt.mode == 3)_vu.keyOnSampler(ev.channel, ev.value0, ev.value1);
				else _vu.keyOn(ev.channel, ev.value0, ev.value1);
				return;
			}
		}
	}
	@Override
	protected void _readSysExData(MidiEvent ev)
	{
		final byte[] _data = ev.data;
		final int _dataEnd = ev.dataSize - 1;  // minus F7
		if (_dataEnd < 8)return;
		int p = 0, v = 0;
		if ((_data[p++]&0xff) != _manufacID) return;
		if ((_data[p++]&0xff) != _deiceID) return;
		if ((_data[p++]&0xff) != _modelID) return;
		int b = _data[p++]&0xff;
		switch (b)	// Address 1
		{
		// Common
			case 0x00:
				b = (_data[p++]&0xff)<<8 | (_data[p++]&0xff);
				switch (b)	// Address 2,3
				{
					case 0x0000 : setMasterVolume(_masterVolumeTable[(_data[p++]&127)]); if (p >= _dataEnd) break;
					case 0x0001 : setReferencePitch(440.0f + ((_data[p++]&127) - 64) * 0.5f); break;
				}
				break;
		// Part
			case 0x02:
				v = _data[p++] & 15;	// Address 2 [ CH ]
				b = _data[p++]&0x7f;
				switch (b)	// Address 3
				{
					case 0x00 : setDefaultPartMode(v, _data[p++]); if (p >= _dataEnd) break;
					case 0x01 : setPartMode(v, _data[p++]); break;
				}
				break;
		// Inst
			case 0x04:
				TonePresetPCM _preset = tonePresets[_data[p++]&127];	// Address 2 [ ID ]
				b = (_data[p++]&0xff);
				switch (b)	// Address 3
				{
					case 0x00 : _preset.toneName[0]  = (char)_data[p++]; if (p >= _dataEnd) break;
					case 0x01 : _preset.toneName[1]  = (char)_data[p++]; if (p >= _dataEnd) break;
					case 0x02 : _preset.toneName[2]  = (char)_data[p++]; if (p >= _dataEnd) break;
					case 0x03 : _preset.toneName[3]  = (char)_data[p++]; if (p >= _dataEnd) break;
					case 0x04 : _preset.toneName[4]  = (char)_data[p++]; if (p >= _dataEnd) break;
					case 0x05 : _preset.toneName[5]  = (char)_data[p++]; if (p >= _dataEnd) break;
					case 0x06 : _preset.toneName[6]  = (char)_data[p++]; if (p >= _dataEnd) break;
					case 0x07 : _preset.toneName[7]  = (char)_data[p++]; if (p >= _dataEnd) break;
					case 0x08 : _preset.toneName[8]  = (char)_data[p++]; if (p >= _dataEnd) break;
					case 0x09 : _preset.toneName[9]  = (char)_data[p++]; if (p >= _dataEnd) break;
					case 0x0A : _preset.toneName[10] = (char)_data[p++]; if (p >= _dataEnd) break;
					case 0x0B : _preset.toneName[11] = (char)_data[p++]; if (p >= _dataEnd) break;
					case 0x0C : _preset.toneName[12] = (char)_data[p++]; if (p >= _dataEnd) break;
					case 0x0D : _preset.toneName[13] = (char)_data[p++]; if (p >= _dataEnd) break;
					case 0x0E : _preset.toneName[14] = (char)_data[p++]; if (p >= _dataEnd) break;
					case 0x0F : _preset.toneName[15] = (char)_data[p++]; if (p >= _dataEnd) break;

					case 0x10 : p++; if (p >= _dataEnd) break;
					case 0x11 : p++; if (p >= _dataEnd) break;
					case 0x12 : _preset.lfoFreq = _preset.lfoFreq & 0x3f80 | (_data[p++]&127);      if (p >= _dataEnd) break;
					case 0x13 : _preset.lfoFreq = _preset.lfoFreq & 0x007f | (_data[p++]&127) << 7; if (p >= _dataEnd) break;
					case 0x14 :_preset.strc     = _data[p++] & 3;   if (p >= _dataEnd) break;
					case 0x15 : _preset.tune    = _preset.tune & 0x3f80 | (_data[p++]&127);      if (p >= _dataEnd) break;
					case 0x16 : _preset.tune    = _preset.tune & 0x007f | (_data[p++]&127) << 7; if (p >= _dataEnd) break;
					case 0x17 : _preset.lms     = _data[p++] & 15;  if (p >= _dataEnd) break;
					case 0x18 : _preset.smplNum = _data[p++] & 127; if (p >= _dataEnd) break;
					case 0x19 : _preset.tl      = _data[p++] & 127; if (p >= _dataEnd) break;
					case 0x1A : _preset.ar      = _data[p++] & 127; if (p >= _dataEnd) break;
					case 0x1B : _preset.dr      = _data[p++] & 127; if (p >= _dataEnd) break;
					case 0x1C : _preset.sl      = _data[p++] & 127; if (p >= _dataEnd) break;
					case 0x1D : _preset.rr      = _data[p++] & 127; if (p >= _dataEnd) break;
					case 0x1E : _preset.kf      = _data[p++] & 3;   if (p >= _dataEnd) break;
					case 0x1F : _preset.kfStart = _data[p++] & 127; break;
				}
				break;
		// Percussion
			case 0x06:
				SamplerPreset _smplPrst = samplerPresets[(_data[p++]&127)];	// Address 2 [ ID ]
				b = (_data[p++]&0xff);
				switch (b)	// Address 3
				{
					case 0x00 : _smplPrst.toneName[0]  = (char)_data[p++]; if (p >= _dataEnd) break;
					case 0x01 : _smplPrst.toneName[1]  = (char)_data[p++]; if (p >= _dataEnd) break;
					case 0x02 : _smplPrst.toneName[2]  = (char)_data[p++]; if (p >= _dataEnd) break;
					case 0x03 : _smplPrst.toneName[3]  = (char)_data[p++]; if (p >= _dataEnd) break;
					case 0x04 : _smplPrst.toneName[4]  = (char)_data[p++]; if (p >= _dataEnd) break;
					case 0x05 : _smplPrst.toneName[5]  = (char)_data[p++]; if (p >= _dataEnd) break;
					case 0x06 : _smplPrst.toneName[6]  = (char)_data[p++]; if (p >= _dataEnd) break;
					case 0x07 : _smplPrst.toneName[7]  = (char)_data[p++]; if (p >= _dataEnd) break;
					case 0x08 : _smplPrst.toneName[8]  = (char)_data[p++]; if (p >= _dataEnd) break;
					case 0x09 : _smplPrst.toneName[9]  = (char)_data[p++]; if (p >= _dataEnd) break;
					case 0x0A : _smplPrst.toneName[10] = (char)_data[p++]; if (p >= _dataEnd) break;
					case 0x0B : _smplPrst.toneName[11] = (char)_data[p++]; if (p >= _dataEnd) break;
					case 0x0C : _smplPrst.toneName[12] = (char)_data[p++]; if (p >= _dataEnd) break;
					case 0x0D : _smplPrst.toneName[13] = (char)_data[p++]; if (p >= _dataEnd) break;
					case 0x0E : _smplPrst.toneName[14] = (char)_data[p++]; if (p >= _dataEnd) break;
					case 0x0F : _smplPrst.toneName[15] = (char)_data[p++]; if (p >= _dataEnd) break;
				
					case 0x10 : _smplPrst.smplNum = _data[p++] & 127; if (p >= _dataEnd) break;
					case 0x11 : _smplPrst.pan     = _data[p++] & 127; if (p >= _dataEnd) break;
					case 0x12 : _smplPrst.octv    = _data[p++] & 15;  if (p >= _dataEnd) break;
					case 0x13 : _smplPrst.tune    = _smplPrst.tune & 0x3f80 | (_data[p++]&127);      if (p >= _dataEnd) break;
					case 0x14 : _smplPrst.tune    = _smplPrst.tune & 0x007f | (_data[p++]&127) << 7; if (p >= _dataEnd) break;
					case 0x15 : _smplPrst.tl      = _data[p++] & 127; if (p >= _dataEnd) break;
					case 0x16 : _smplPrst.ar      = _data[p++] & 127; if (p >= _dataEnd) break;
					case 0x17 : _smplPrst.dr      = _data[p++] & 127; if (p >= _dataEnd) break;
					case 0x18 : _smplPrst.sl      = _data[p++] & 127; if (p >= _dataEnd) break;
					case 0x19 : _smplPrst.rr      = _data[p++] & 127; break;
				}
				break;
		}
	}
}