/*! ********************************************
 * SynthesizerWT.java
 * Copyright 2019 masahiro takamoto (MotChan)
 * Distributed under the MIT license. 
 * see -> sora/license.txt
 ******************************************** */
package sora.graybox.synth;

import sora.graybox.MidiEvent;
// WaveTable Synthesis
public final class SynthesizerWT extends MultiTimberSynthBase
{
	static final int WAVET_CONT_SIZE_BITS = 7;
	static final int WAVET_CONT_SIZE      = 1 << WAVET_CONT_SIZE_BITS;
	static final int PG_PHASE_FP = 18;  // 18bit fixed-point  (precision 1 / 262144)

private class VoiceUnitWT extends VoiceUnitBase
{
// Tables
	// Log2 Volume Table
	private float[] _LOG2V_TABLE = null;
	// Pow2 Table
	private final int _POW2T_OCTV_BITS  = SynthData.POW2T_OCTV_BITS;
	private final int _POW2T_OCTV_MASK  = SynthData.POW2T_OCTV_MASK;
	private int[] _POW2_TABLE      = null;
	private int[] _POW2T_RS_TABLE  = null;
	private int[] _POW2T_KEY_TABLE = null;
	// EG
	private final int _EG_PRAM_OCTV_LS = SynthData.POW2T_OCTV_BITS - 3;
	private final int _EG_CNT_FP       = SynthData.EG_CNT_FP;
	private final int _EG_MAX_CNT      = SynthData.EG_MAX_CNT;
	private final int _EG_CNT_ENV_ATK  = SynthData.EG_CNT_ENV_ATK;
	private final int _EG_CNT_ENV_BTM  = SynthData.EG_CNT_ENV_BTM;
	private final int _EG_SUSTAIN_BTM  = SynthData.EG_SUSTAIN_BTM;
	private int[] _EG_TOTAL_LEVEL_TABLE = null;
	private int[] _EG_LEVEL_TABLE_UP    = null;
	private int[] _EG_LEVEL_TABLE_DW    = null;
	private int[] _EG_KEY_TABLE         = null;
	private int[] _EG_POW2T_RS_TABLE    = null;
	private int[] _EG_CNT_GATE_TABLE    = null;
	// PG
	private final int _PG_PHASE_FP      = SynthesizerWT.PG_PHASE_FP;
	private final int _PG_PHASE_FP_MASK = (1 << _PG_PHASE_FP) - 1;
	// LFO
	private final int _LFO_TIMER_INIT_VALUE = SynthData.LFO_TIMER_INIT_VALUE;
	private final int _LFO_WAVE_TABLE_MASK  = SynthData.LFO_WAVE_TABLE_SIZE - 1;
	private int[] _LFO_WAVE_TABLE = null;
	// Noise Generator (15bit LFSR
	private final int _LFSR_TAP_15 = 0x6000;
	private int _lfsrReg = 1;
	private int _noise   = 0;
// Synthesizer Data
	VoicePart[] VOICE_PARTS = null;
	int[][] WAVE_TABLES = null;
	TonePresetWT[] TONE_PRESETS = null;
	SampleData[] SAMPLE_DATAS = null;
	SamplerPreset[] SAMPLER_PRESETS = null;
	
	int pgPhaseMask      = 0;
	int pgPhaseMaskNoise = 0;
	int pgSamplePow2tIdx = 0;  // PG PhaseStep (Pow2Table Index Offset) - SamplingRate
	int pgKeyPow2tIdx    = 0;  // PG PhaseStep (Pow2Table Index Offset) - ReferenceKey (NoteNumber 0)
	int egPow2tIdx       = 0;  // EG PhaseStep (Pow2Table Index Offset)
	int egPhaseStepSndOff= 0;
	int lfoTimerCoef     = 0;  // Coefficient Value - LFO Phase Update Timer
// VoiceUnit Status
	// Pitch
	private int _ptBend     = 0;   // pitchBend
	private int _ptBendSens = 0;   // pitchBend Sensitivity
	// LFO
	private int _lfoPhase     = 0;  // LFO Phase
	private int _lfoTimer     = 0;  // LFO Phase Update Timer
	private int _lfoTimerStep = 0;  // Update Timer Step
	private int _lfoMod       = 0;  // Modulation Depth
	private int _lfoModSens   = 0;  // Modulation Sensitivity
	private int _lfoOutPM     = 0;  // output - phase modulation
	private int _lfoOutAM     = 0;  // output - amplitude modulation
	// Wave
	private int[] _waveTable = null;
	// EG
	/*
	- Envelop Generator State -
	[ 5674 3210 ] 8Bits Flag Status (0 = deActive)
	[ ---- RSDA ] "-" = Reserved Bits
		A = Attack  stage
		D = Decay   stage
		S = Sustain stage
		R = Release stage
	*/
	private int _egState      = 0;
	private int[] _egLevelTable = null;
	private int _egCount      = 0;
	private int _egCountStep  = 0;
	//private int _egCountStepAR= 0;
	private int _egCountStepDR= 0;
	private int _egCountStepRR= 0;
	private int _egCntEnvSus  = 0;
	private int _egCntEnvBtm  = 0;
	private int _egGate       = 0;
	private int _egSL         = 0;
	private int _egTL         = 0;
	private int _egTLv        = 0;
	// PG
	private int _pgKey       = 0;
	private int _pgBend      = 0;
	private int _pgPhase     = 0;
	private int _pgPhaseStep = 0;
// Sampler
	private int _pgPhaseAFP = 0;  // PG Phase (After the Fixed Point)
	private int _plbkE      = 0;  // PlayBack End Frame   [plbkE <= channelPerSamples]
	private int _loopE      = 0;  // Loop End Frame       [loopE <= plbkE]
	private int _loopF      = 0;  // Loop Frames          [val = loopE - lpStart]
	private int _loopFk     = 0;  // Loop Frames          * after -> keyOff
	
	@Override
	void keyOff()
	{
		// KeyOn --->
		if ((state & 0x0010) != 0)
		{
			// ATTACK ->
			if (_egState == 1)
			{
				_egLevelTable = _EG_LEVEL_TABLE_DW;
				if (_egCount >= _EG_MAX_CNT)_egCount = 0;
				else _egCount = _EG_LEVEL_TABLE_UP[_egCount >> _EG_CNT_FP] << _EG_CNT_FP - 1;  // -1 = 8192 to 4096
				if (_egCount > _egCntEnvBtm){ killProcess(); return; }
			}
			// DECAY,SUSTAIN ->
			_egGate = _egCntEnvBtm;
			_egCountStep = _egCountStepRR;
			_egState = 8;  // Release
			state = 0x0021;
			// sampler ---
			_loopE = _plbkE;
			_loopF = _loopFk;
		}
	}
	@Override
	void soundOff()
	{
		// KeyOn || KeyOff --->
		if ((state & 0x0030)!= 0)
		{
			// ATTACK ->
			if (_egState == 1)
			{
				_egLevelTable = _EG_LEVEL_TABLE_DW;
				if (_egCount >= _EG_MAX_CNT)_egCount = 0;
				else _egCount = _EG_LEVEL_TABLE_UP[_egCount >> _EG_CNT_FP] << _EG_CNT_FP - 1;
				if (_egCount > _egCntEnvBtm){ killProcess(); return; }
			}
			// DECAY,SUSTAIN ->
			_egGate = _egCntEnvBtm;
			_egCountStep = egPhaseStepSndOff;
			_egState = 0;  // SoundOff, EG OFF
			partNumber = -1; // SoundOff
			state = 0x0041;
			// sampler ---
			_loopE = _plbkE;
			_loopF = _loopFk;
		}
	}
	private int _shiftEGState()
	{
		switch (_egState)
		{
			// ATTACK ->
			case 1:
				_egLevelTable = _EG_LEVEL_TABLE_DW;
				_egCount = 0;
				// -> DECAY
				if (_egSL > 0)
				{
					_egGate = _egCntEnvSus;
					_egCountStep = _egCountStepDR;
					return _egState = 2;
				}
				// -> SUSTAIN
				_egGate = 0x7fffffff;
				_egCountStep = 0;
				return _egState = 4;
			// DECAY ->
			case 2:
				// -> SUSTAIN
				if ((_egTL + _egSL) < _EG_SUSTAIN_BTM)
				{
					_egLevelTable = _EG_LEVEL_TABLE_DW;
					_egGate = 0x7fffffff;
					_egCountStep = 0;
					_egCount = _egCntEnvSus;
					return _egState = 4;
				}
			// DECAY,RELEASE,SoundOFF -> EG OFF
			default: return _egState = 0;
		}
	}
	@Override
	void keyOn(int ch, int note, int velo)
	{
	// KeyOn Status ---
		state = 0x0011;
		partNumber = ch;
		noteNumber = note;
	// Part, Preset ---
		VoicePart _part = VOICE_PARTS[partNumber];
		TonePresetWT _preset;
		int _keyFlw = egPow2tIdx;
		if (_part.mode == 2)
		{
			_preset = TONE_PRESETS[noteNumber | 128];
			_pgKey = _POW2T_KEY_TABLE[_preset.key] + (_preset.tune - 8192) + pgKeyPow2tIdx;
			//_keyFlw = 0;
			_prPan = _preset.pan;
		}
		else
		{
			_preset = TONE_PRESETS[_part.programChange];
			_pgKey = _POW2T_KEY_TABLE[noteNumber] + (_preset.tune - 8192) + pgKeyPow2tIdx;
			_keyFlw += _EG_KEY_TABLE[127 + noteNumber - _preset.kfStart] << _preset.kf;
			_prPan = 64;
		}
	// wave ---
		_waveTable = WAVE_TABLES[_preset.wave];
	// LFO ---
		_lfoModSens = _preset.lms;
		_lfoMod = _part.ccValues[1] * _lfoModSens;
		_lfoTimerStep = _preset.lfoFreq * lfoTimerCoef;
		_lfoTimer = _LFO_TIMER_INIT_VALUE;
		_lfoPhase = _preset.strc + 1 & 1;
		_lfoOutPM = 0;
		_lfoOutAM = 0;
	// volume ---
		_vPtVol = _mainVolTable[_part.ccValues[7]];
		_vPtExp = _exprVolTable[_part.ccValues[11]];
		_vVel = _veloVolTable[velo];
		_vVol = _vVel * _masterVolume;
		_vOut = _vPtVol * _vPtExp * _vVol;
		_vPanL = _panTable[      _prPan + _part.ccValues[10]];
		_vPanR = _panTable[256 - _prPan - _part.ccValues[10]];
		_vOutL = _vOut * _vPanL;
		_vOutR = _vOut * _vPanR;
	// PG ---
		_ptBendSens = _POW2T_KEY_TABLE[_part.bendSens];
		_ptBend = _part.pitchBend - 8192;
		_pgBend = _ptBend * _ptBendSens >> 13;
		int v = _pgKey + _pgBend;// & 0x7fffffff;
		_pgPhaseStep = _POW2_TABLE[v & _POW2T_OCTV_MASK] >> _POW2T_RS_TABLE[v >> _POW2T_OCTV_BITS];
		_pgPhase = 0;
	// EG ---
		_egTL = _preset.tl;
		_egTLv = _EG_TOTAL_LEVEL_TABLE[_egTL];
		_egCntEnvBtm = _EG_CNT_GATE_TABLE[128 - _egTL];
		_egSL = _preset.sl;
		_egCntEnvSus = _EG_CNT_GATE_TABLE[_egSL];
		if (_egCntEnvSus > _egCntEnvBtm)_egCntEnvSus = _egCntEnvBtm;
		v = (_preset.dr << _EG_PRAM_OCTV_LS) + _keyFlw;
		_egCountStepDR = _POW2_TABLE[v & _POW2T_OCTV_MASK] >> _EG_POW2T_RS_TABLE[v >> _POW2T_OCTV_BITS];
		v = (_preset.rr << _EG_PRAM_OCTV_LS) + _keyFlw;
		_egCountStepRR = _POW2_TABLE[v & _POW2T_OCTV_MASK] >> _EG_POW2T_RS_TABLE[v >> _POW2T_OCTV_BITS];
	// EG State ---
		// -> ATTACK
		if (_preset.ar < 127) {
			_egCount = _EG_CNT_ENV_ATK;
			_egLevelTable = _EG_LEVEL_TABLE_UP;
			_egGate = _EG_MAX_CNT;
			v = (_preset.ar << _EG_PRAM_OCTV_LS) + _keyFlw;
			_egCountStep = _POW2_TABLE[v & _POW2T_OCTV_MASK] >> _EG_POW2T_RS_TABLE[v >> _POW2T_OCTV_BITS];
			_egState = 1;
		} else {
			_egLevelTable = _EG_LEVEL_TABLE_DW;
			_egCount = 0;
		// -> DECAY
			if (_egSL > 0) {
				_egGate = _egCntEnvSus;
				_egCountStep = _egCountStepDR;
				_egState = 2;
		// -> SUSTAIN
			} else {
				_egGate = 0x7fffffff;  // dummy
				_egCountStep = 0;  // dummy
				_egState = 4; }
		}
	// process ---
		process = _PROC_PTRS[_preset.strc];
	}
	void keyOnSampler(int ch, int note, int velo)
	{
	// KeyOn Status ---
		state = 0x0011;
		partNumber = ch;
		noteNumber = note;
	// Part, Preset ---
		VoicePart _part = VOICE_PARTS[partNumber];
		SamplerPreset _preset = SAMPLER_PRESETS[noteNumber];
		SampleData _sample = SAMPLE_DATAS[_preset.smplNum];
		if (_sample.sampleData == null){ killProcess(); return; }
		_waveTable = _sample.sampleData;
		_plbkE = _sample.plbkE;
		_loopE = _sample.loopE;
		_loopF = _sample.loopF;
		_loopFk = _sample.loopFk;
	// LFO ---
		_lfoOutPM = 0;
		_lfoOutAM = 0;
	// volume ---
		_vPtVol = _mainVolTable[_part.ccValues[7]];
		_vPtExp = _exprVolTable[_part.ccValues[11]];
		_vVel = _veloVolTable[velo];
		_vVol = _vVel * _masterVolume;
		_vOut = _vPtVol * _vPtExp * _vVol;
		_prPan = _preset.pan;
		_vPanL = _panTable[      _prPan + _part.ccValues[10]];
		_vPanR = _panTable[256 - _prPan - _part.ccValues[10]];
		_vOutL = _vOut * _vPanL;
		_vOutR = _vOut * _vPanR;
	// PG ---
		_ptBendSens = _POW2T_KEY_TABLE[_part.bendSens];
		_ptBend = _part.pitchBend - 8192;
		_pgBend = _ptBend * _ptBendSens >> 13;
		_pgKey = (_preset.octv - 8 << 13) + (_preset.tune - 8192) + _sample.pgSample + pgSamplePow2tIdx;
		int v = _pgKey + _pgBend;
		_pgPhaseStep = _POW2_TABLE[v & _POW2T_OCTV_MASK] >> _POW2T_RS_TABLE[v >> _POW2T_OCTV_BITS];
		_pgPhase = _sample.plbkS;
		_pgPhaseAFP = 0;
	// EG ---
		_egTL = _preset.tl;
		_egTLv = _EG_TOTAL_LEVEL_TABLE[_egTL];
		_egCntEnvBtm = _EG_CNT_GATE_TABLE[128 - _egTL];
		_egSL = _preset.sl;
		_egCntEnvSus = _EG_CNT_GATE_TABLE[_egSL];
		if (_egCntEnvSus > _egCntEnvBtm)_egCntEnvSus = _egCntEnvBtm;
		v = (_preset.dr << _EG_PRAM_OCTV_LS) + egPow2tIdx;
		_egCountStepDR = _POW2_TABLE[v & _POW2T_OCTV_MASK] >> _EG_POW2T_RS_TABLE[v >> _POW2T_OCTV_BITS];
		v = (_preset.rr << _EG_PRAM_OCTV_LS) + egPow2tIdx;
		_egCountStepRR = _POW2_TABLE[v & _POW2T_OCTV_MASK] >> _EG_POW2T_RS_TABLE[v >> _POW2T_OCTV_BITS];
	// EG State ---
		// -> ATTACK
		if (_preset.ar < 127) {
			_egCount = _EG_CNT_ENV_ATK;
			_egLevelTable = _EG_LEVEL_TABLE_UP;
			_egGate = _EG_MAX_CNT;
			v = (_preset.ar << _EG_PRAM_OCTV_LS) + egPow2tIdx;
			_egCountStep = _POW2_TABLE[v & _POW2T_OCTV_MASK] >> _EG_POW2T_RS_TABLE[v >> _POW2T_OCTV_BITS];
			_egState = 1;
		} else {
			_egLevelTable = _EG_LEVEL_TABLE_DW;
			_egCount = 0;
		// -> DECAY
			if (_egSL > 0) {
				_egGate = _egCntEnvSus;
				_egCountStep = _egCountStepDR;
				_egState = 2;
		// -> SUSTAIN
			} else {
				_egGate = 0x7fffffff;  // dummy
				_egCountStep = 0;  // dummy
				_egState = 4; }
		}
	// process ---
		process = (_sample.channelCount == 2)? this::_procSmpl2ch: this::_procSmpl1ch;
	}
	@Override
	void updPartBendSens(int val)
	{
		_ptBendSens = _POW2T_KEY_TABLE[val];
		_pgBend = _ptBend * _ptBendSens >> 13;
		int v = _pgKey + _pgBend + _lfoOutPM;
		_pgPhaseStep = _POW2_TABLE[v & _POW2T_OCTV_MASK] >> _POW2T_RS_TABLE[v >> _POW2T_OCTV_BITS];
	}
	@Override
	void updPartPitchBend(int val)
	{
		_ptBend = val - 8192;
		_pgBend = _ptBend * _ptBendSens >> 13;
		int v = _pgKey + _pgBend + _lfoOutPM;
		_pgPhaseStep = _POW2_TABLE[v & _POW2T_OCTV_MASK] >> _POW2T_RS_TABLE[v >> _POW2T_OCTV_BITS];
	}
	@Override
	void updPartModulation(int val)
	{
		_lfoMod = val * _lfoModSens;
	}
	void updateLfoFrequency(int val)
	{
		_lfoTimerStep = val * lfoTimerCoef;
	}
	void setup(SynthData synthData)
	{
		// volume Table
		_mainVolTable = synthData.getVOLUME_TABLE_0dB();
		_exprVolTable = synthData.getVOLUME_TABLE_0dB();
		_veloVolTable = synthData.getVOLUME_TABLE_0dB();
		_panTable     = synthData.getPAN_TABLE_SIN();
		// Log2 Volume Table
		_LOG2V_TABLE = synthData.getLOG2V_TABLE();
		// Pow2 Table
		_POW2_TABLE      = synthData.getPOW2_TABLE();
		_POW2T_RS_TABLE  = synthData.getPOW2T_RS_TABLE();
		_POW2T_KEY_TABLE = synthData.getPOW2T_KEY_TABLE();
		// EG
		_EG_TOTAL_LEVEL_TABLE = synthData.getEG_TOTAL_LEVEL_TABLE();
		_EG_LEVEL_TABLE_UP    = synthData.getEG_LEVEL_TABLE_UP();
		_EG_LEVEL_TABLE_DW    = synthData.getEG_LEVEL_TABLE_DW();
		_EG_KEY_TABLE         = synthData.getEG_KEY_TABLE();
		_EG_POW2T_RS_TABLE    = synthData.getEG_POW2T_RS_TABLE();
		_EG_CNT_GATE_TABLE    = synthData.getEG_CNT_GATE_TABLE();
		// etc
		_LFO_WAVE_TABLE = synthData.getLFO_WAVE_TABLES()[0];
	}
// Constructor
	VoiceUnitWT()
	{
		super();
		_PROC_PTRS[0] = this::_proc0;
		_PROC_PTRS[1] = this::_proc1;
		_PROC_PTRS[2] = this::_proc2;
		_PROC_PTRS[3] = this::_proc3;
	}
// process ---
	// pointer array (process Function)
	private final Function_PRC[] _PROC_PTRS = new Function_PRC[4];
	// Inst
	private void _proc0(float[] outL, float[] outR, int end)
	{
		int[]  _et = _egLevelTable;
		final float[] _log2vT = _LOG2V_TABLE;
		final int[] _wt = _waveTable;
		int i = startFrame;
		for ( ; i < end; i++)
		{
		// eg update --->
			if (_egCount >= _egGate)
			{
				if (_shiftEGState() != 0)_et = _egLevelTable;
				else { killProcess(); return; }
			}
			int v = _et[_egCount >> _EG_CNT_FP] + _egTLv;
		// pg update --->
			v = _wt[_pgPhase >> _PG_PHASE_FP] + v;
			float n = _log2vT[v];
		// out --->
			outL[i] += n * _vOutL;
			outR[i] += n * _vOutR;
		// counter --->
			_egCount += _egCountStep;
			_pgPhase = _pgPhase + _pgPhaseStep & pgPhaseMask;
		}
		startFrame = i;
	}
	// Inst + LFO PM
	private void _proc1(float[] outL, float[] outR, int end)
	{
		int[]  _et = _egLevelTable;
		final float[] _log2vT = _LOG2V_TABLE;
		final int[] _wt = _waveTable;
		int i = startFrame;
		for ( ; i < end; i++)
		{
		// lfo Update --->
			if (_lfoTimer <= 0)
			{
				_lfoOutPM = (_LFO_WAVE_TABLE[_lfoPhase] * _lfoMod) >> 11;  // 7(mod) + 4(sens)
				int l = _pgKey + _pgBend + _lfoOutPM;
				_pgPhaseStep = _POW2_TABLE[l & _POW2T_OCTV_MASK] >> _POW2T_RS_TABLE[l >> _POW2T_OCTV_BITS];
				_lfoPhase = _lfoPhase + 2 & _LFO_WAVE_TABLE_MASK;
				_lfoTimer += _LFO_TIMER_INIT_VALUE;
			}
		// eg update --->
			if (_egCount >= _egGate)
			{
				if (_shiftEGState() != 0)_et = _egLevelTable;
				else { killProcess(); return; }
			}
			int v = _et[_egCount >> _EG_CNT_FP] + _egTLv;
		// pg update --->
			v = _wt[_pgPhase >> _PG_PHASE_FP] + v;
			float n = _log2vT[v];
		// out --->
			outL[i] += n * _vOutL;
			outR[i] += n * _vOutR;
		// counter --->
			_lfoTimer -= _lfoTimerStep;
			_egCount += _egCountStep;
			_pgPhase = _pgPhase + _pgPhaseStep & pgPhaseMask;
		}
		startFrame = i;
	}
	// Inst + LFO AM
	private void _proc2(float[] outL, float[] outR, int end)
	{
		int[]  _et = _egLevelTable;
		final float[] _log2vT = _LOG2V_TABLE;
		final int[] _wt = _waveTable;
		int i = startFrame;
		for ( ; i < end; i++)
		{
		// lfo Update --->
			if (_lfoTimer <= 0)
			{
				int l = (_LFO_WAVE_TABLE[_lfoPhase] * _lfoMod) >> 11;
				_lfoOutAM = l & 0x7ffffffe;
				_lfoPhase = _lfoPhase + 2 & _LFO_WAVE_TABLE_MASK;
				_lfoTimer += _LFO_TIMER_INIT_VALUE;
			}
		// eg update --->
			if (_egCount >= _egGate)
			{
				if (_shiftEGState() != 0)_et = _egLevelTable;
				else { killProcess(); return; }
			}
			int v = _et[_egCount >> _EG_CNT_FP] + _egTLv;
		// pg update --->
			v = _wt[_pgPhase >> _PG_PHASE_FP] + v + _lfoOutAM;
			float n = _log2vT[v];
		// out --->
			outL[i] += n * _vOutL;
			outR[i] += n * _vOutR;
		// counter --->
			_lfoTimer -= _lfoTimerStep;
			_egCount += _egCountStep;
			_pgPhase = _pgPhase + _pgPhaseStep & pgPhaseMask;
		}
		startFrame = i;
	}
	// Noise
	private void _proc3(float[] outL, float[] outR, int end)
	{
		int[]  _et = _egLevelTable;
		final float[] _log2vT = _LOG2V_TABLE;
		int i = startFrame;
		for ( ; i < end; i++)
		{
		// eg update --->
			if (_egCount >= _egGate)
			{
				if (_shiftEGState() != 0)_et = _egLevelTable;
				else { killProcess(); return; }
			}
			int v = _et[_egCount >> _EG_CNT_FP] + _egTLv;
		// pg update --->
			//while (_pgPhase > _PG_PHASE_FP_MASK)
			if (_pgPhase > pgPhaseMaskNoise)
			{
				//_pgPhase -= 1 << _PG_PHASE_FP;
				_pgPhase &= pgPhaseMaskNoise;
				_lfsrReg = _lfsrReg >> 1 ^ -(_lfsrReg & 1) & _LFSR_TAP_15;
				_noise = _lfsrReg & 1;
			}
			float n = _log2vT[_noise + v];
		// out --->
			outL[i] += n * _vOutL;
			outR[i] += n * _vOutR;
		// counter --->
			_egCount += _egCountStep;
			_pgPhase += _pgPhaseStep;
		}
		startFrame = i;
	}
	// Sampler 1ch
	private void _procSmpl1ch(float[] outL, float[] outR, int end)
	{
		int[]  _et = _egLevelTable;
		final float[] _log2vT = _LOG2V_TABLE;
		final int[] _wt = _waveTable;
		int i = startFrame;
		for ( ; i < end; i++)
		{
		// eg update --->
			if (_egCount >= _egGate)
			{
				if (_shiftEGState() != 0)_et = _egLevelTable;
				else { killProcess(); return; }
			}
			int v = _et[_egCount >> _EG_CNT_FP] + _egTLv;	
		// pg update --->
			if (_pgPhase >= _loopE)
			{
				if (_loopF > 0) do{ _pgPhase -= _loopF; } while (_pgPhase >= _loopE);
				else { killProcess(); return; }
			}
			v = _wt[_pgPhase] + v;
			float n = _log2vT[v];
		// out --->
			outL[i] += n * _vOutL;
			outR[i] += n * _vOutR;
		// counter --->
			_egCount += _egCountStep;
			_pgPhaseAFP += _pgPhaseStep;
			_pgPhase += _pgPhaseAFP >> _PG_PHASE_FP;
			_pgPhaseAFP &= _PG_PHASE_FP_MASK;
		}
		startFrame = i;
	}
	// Sampler 2ch
	private void _procSmpl2ch(float[] outL, float[] outR, int end)
	{
		int[]  _et = _egLevelTable;
		final float[] _log2vT = _LOG2V_TABLE;
		final int[] _wt = _waveTable;
		int i = startFrame;
		for ( ; i < end; i++)
		{
		// eg update --->
			if (_egCount >= _egGate)
			{
				if (_shiftEGState() != 0)_et = _egLevelTable;
				else { killProcess(); return; }
			}
			int v = _et[_egCount >> _EG_CNT_FP] + _egTLv;	
		// pg update --->
			if (_pgPhase >= _loopE)
			{
				if (_loopF > 0) do{ _pgPhase -= _loopF; } while (_pgPhase >= _loopE);
				else { killProcess(); return; }
			}
		// out --->
			int p = _pgPhase << 1;
			outL[i] += _log2vT[_wt[p    ] + v] * _vOutL;
			outR[i] += _log2vT[_wt[p + 1] + v] * _vOutR;
		// counter --->
			_egCount += _egCountStep;
			_pgPhaseAFP += _pgPhaseStep;
			_pgPhase += _pgPhaseAFP >> _PG_PHASE_FP;
			_pgPhaseAFP &= _PG_PHASE_FP_MASK;
		}
		startFrame = i;
	}
}

// Factory -----
	private void _resetIndexTable()
	{
		int _size = 1 << _maxWaveValueBits, _half = _size >> 1;
		if (_indexTable == null)_indexTable = new int[_size];
		final double _LOGE = -(SynthData.LOG2VT_DYN_RES * 1.442695040888963387);
		final double _BOADER = 1.0 / (1 << SynthData.LOG2VT_DYN_BITS);
		final int _BOTTOM = SynthData.LOG2VT_SIZE;
		for (int i = 0; i < _size; i++)
		{
			double v = (i - _half) * 0.0078125;
			if      (v >  _BOADER) _indexTable[i] =  (int)(Math.log( v) * _LOGE + 0.5) << 1;
			else if (v < -_BOADER) _indexTable[i] = ((int)(Math.log(-v) * _LOGE + 0.5) << 1) + 1;
			else _indexTable[i] = _BOTTOM;
		}
	}
	@Override
	protected void _recreateVU(int maxPoly)
	{
		// Recreate Array
		_voiceUnits = new VoiceUnitBase[maxPoly];
		_voiceUnitExt = new VoiceUnitWT[maxPoly];
		// Reset VoiceUnit
		for (int i = 0; i < maxPoly; i++)
		{
			_voiceUnitExt[i] = new VoiceUnitWT();
			VoiceUnitWT _vu = _voiceUnitExt[i];
			_voiceUnits[i] = _vu;
			_vu.setup(_synthData);
			_vu.killProcess();
			_vu.VOICE_PARTS = voiceParts;
			_vu.WAVE_TABLES = waveTables;
			_vu.TONE_PRESETS = tonePresets;
			_vu.SAMPLE_DATAS = sampleDatas;
			_vu.SAMPLER_PRESETS = samplerPresets;
			_vu.pgPhaseMask = (_waveTableSize << PG_PHASE_FP) - 1;
			_vu.pgPhaseMaskNoise = _vu.pgPhaseMask >> 5;  // 32/1 update cycle
			_vu.setMasterVolume(_masterVolume);
		}
		// set Property
		_maxPolyphony = maxPoly;
	}
	@Override
	protected void _resetKeyParams()
	{
		final double _sampleRateD = (double)_sampleRate;
		// LFO
		double _lfoWaveSize = SynthData.LFO_WAVE_TABLE_SIZE / 2.0;
			// Max Frame Rate - LFO Phase Step
		double _lfoMaxFrameRate = _sampleRateD / SynthData.LFO_MAX_FREQ / _lfoWaveSize;
		if (_lfoMaxFrameRate < 1.0)_lfoMaxFrameRate = 1.0;  // Max 1frame per 1Phase
		int _lfoTimerCoef =  (int)(SynthData.LFO_TIMER_INIT_VALUE / 16383.0 / _lfoMaxFrameRate + 0.5);
		// EG
		double _rate0Frames = SynthData.EG_REFERENCE_TIME * SynthData.LOG2VT_DYN_BITS * _sampleRateD;  // to -96dB Frames
		int _egPow2tIdx = SynthData.toPow2tIndexValue(SynthData.EG_MAX_CNT / _rate0Frames);
		double _sndOffFrames = SynthData.EG_SOUND_OFF_TIME / SynthData.LOG2VT_DYN_BITS * _sampleRateD;  // to -96dB Frames
		int _egSndOff = (int)(SynthData.EG_MAX_CNT / _sndOffFrames + 0.5);
		// PG
		int _smpl = SynthData.toPow2tIndexValue(_sampleRateD);
		int _pgSamplePow2tIdx = SynthData.toPow2tIndexValue((double)(1 << PG_PHASE_FP)) - _smpl;
		// PG WaveTable
		int _pgKeyPow2tIdx = SynthData.toPow2tIndexValue((double)(_waveTableSize << PG_PHASE_FP)) - _smpl;
		_pgKeyPow2tIdx += SynthData.toPow2tIndexValue(_referencePitch * Math.pow(2.0, -69.0 / 12.0)); // note 0 Pitch
		for (int i = 0; i < _maxPolyphony; i++)
		{
			VoiceUnitWT _vu = _voiceUnitExt[i];
			_vu.killProcess();
			_vu.pgSamplePow2tIdx = _pgSamplePow2tIdx;
			_vu.pgKeyPow2tIdx = _pgKeyPow2tIdx;
			_vu.egPow2tIdx = _egPow2tIdx;
			_vu.egPhaseStepSndOff = _egSndOff;
			_vu.lfoTimerCoef = _lfoTimerCoef;
		}
	}
// Tables -----
	private int[] _indexTable = null;
// Synthesizer - Data -----
	private int[][] waveTables = null; // indexed Values
	private SampleData[] sampleDatas = null;
	public TonePresetWT[] tonePresets = null;  // [0 - 127](ProgramChange), [128 - 255](Percussion Key)
	public SamplerPreset[] samplerPresets = null;
	private int _waveTableSizeBits = 5;
	private int _waveTableSize     = 1 << _waveTableSizeBits;
	private int _maxWaveValueBits  = 8;
	private SynthData _synthData = null;
	public int getWaveTableSize(){ return _waveTableSize; }
	public int setWaveTableSize(int size)
	{
		if (size == _waveTableSize) return _waveTableSize;
		int v = 8, _b = 3, _start = _waveTableSize;
		for ( ; v < WAVET_CONT_SIZE; v <<= 1, _b++) if (v >= size) break;
		_waveTableSizeBits = _b;
		_waveTableSize = v;
		final int _BOTTOM = SynthData.LOG2VT_SIZE;
		for (int i = 0; i < 128; i++)
		{
			int[] _table = waveTables[i];
			for (int j = _start; j < _waveTableSize; j++)_table[j] = _BOTTOM;
		}
		for (int i = 0; i < _maxPolyphony; i++)
		{
			VoiceUnitWT _vu = _voiceUnitExt[i];
			_vu.killProcess();
			_vu.pgPhaseMask = (_waveTableSize << PG_PHASE_FP) - 1;
			_vu.pgPhaseMaskNoise = _vu.pgPhaseMask >> 5;
		}
		_resetKeyParams();
		return _waveTableSize;
	}
	public void getWaveForm(int id, byte[] buffer, int bitDepth)
	{
		int _size = buffer.length;
		if (_size > WAVET_CONT_SIZE)_size = WAVET_CONT_SIZE;
		// ---
		if (bitDepth > 8 || bitDepth < 1) bitDepth = 8;
		final float _center = (float)(1 << (bitDepth - 1));
		final int _max = (1 << bitDepth) - 1;
		int[] _table = waveTables[id];
		final float[] _log2vt = _synthData.getLOG2V_TABLE();
		for (int i = 0; i < _size; i++)
		{
			int v = (int)(_log2vt[_table[i]] * _center + _center + 0.5f);
			if (v > _max) v = _max;
			buffer[i] = (byte)(v & 255);
		}
	}
	public void getWaveForm(int id, int[] buffer, int bitDepth)
	{
		int _size = buffer.length;
		if (_size > WAVET_CONT_SIZE)_size = WAVET_CONT_SIZE;
		// ---
		if (bitDepth > 8 || bitDepth < 1) bitDepth = 8;
		final float _center = (float)(1 << (bitDepth - 1));
		final int _max = (1 << bitDepth) - 1;
		int[] _table = waveTables[id];
		final float[] _log2vt = _synthData.getLOG2V_TABLE();
		for (int i = 0; i < _size; i++)
		{
			int v = (int)(_log2vt[_table[i]] * _center + _center + 0.5f);
			if (v > _max) v = _max;
			buffer[i] = v;
		}
	}
	public void getWaveFormN(int id, float[] buffer)
	{
		int _size = buffer.length;
		if (_size > WAVET_CONT_SIZE)_size = WAVET_CONT_SIZE;
		// ---
		int[] _table = waveTables[id];
		final float[] _log2vt = _synthData.getLOG2V_TABLE();
		for (int i = 0; i < _size; i++) buffer[i] = _log2vt[_table[i]];
	}
	public int getWaveTableValue(int id, int index, int bitDepth)
	{
		if (bitDepth > 8 || bitDepth < 1) bitDepth = 8;
		final float _center = (float)(1 << (bitDepth - 1));
		final int _max = (1 << bitDepth) - 1;
		int v = (int)(_synthData.getLOG2V_TABLE()[waveTables[id][index]] * _center + _center + 0.5f);
		if (v > _max) v = _max;
		return v;
	}
	public float getWaveTableValueN(int id, int index)
	{
		return _synthData.getLOG2V_TABLE()[waveTables[id][index]];
	}
	// buffer [0-255(128 = 0)](bitDepth = 8)
	public void setWaveForm(int id, byte[] buffer, int bitDepth)
	{
		int _size = buffer.length;
		if (_size > WAVET_CONT_SIZE)_size = WAVET_CONT_SIZE;
		// ---
		if (bitDepth > 8 || bitDepth < 1) bitDepth = 8;
		final int _shift = 8 - bitDepth;
		int[] _table = waveTables[id];
		for (int i = 0; i < _size; i++)
		{
			int v = (buffer[i] << _shift) & 255;
			_table[i] = _indexTable[v];
		}
	}
	public void setWaveForm(int id, int[] buffer, int bitDepth)
	{
		int _size = buffer.length;
		if (_size > WAVET_CONT_SIZE)_size = WAVET_CONT_SIZE;
		// ---
		if (bitDepth > 8 || bitDepth < 1) bitDepth = 8;
		final int _shift = 8 - bitDepth;
		int[] _table = waveTables[id];
		for (int i = 0; i < _size; i++)
		{
			int v = buffer[i] << _shift;
			_table[i] = _indexTable[v & 0xff];
		}
	}
	// buffer [-1.0 ~ 1.0]
	public void setWaveFormN(int id, float[] buffer){ setWaveFormN(id, buffer, 8); }
	public void setWaveFormN(int id, float[] buffer, int bitDepth)
	{
		int _size = buffer.length;
		if (_size > WAVET_CONT_SIZE)_size = WAVET_CONT_SIZE;
		// ---
		if (bitDepth > 8 || bitDepth < 1) bitDepth = 8;
		final int _shift = 8 - bitDepth;
		int[] _table = waveTables[id];
		for (int i = 0; i < _size; i++)
		{
			int v = (int)(buffer[i] * 127.0f) + 128;
			v >>= _shift;
			_table[i] = _indexTable[v];
		}
	}
	public void setWaveTableValue(int id, int index, byte value, int bitDepth)
	{
		if (bitDepth > 8 || bitDepth < 1) bitDepth = 8;
		waveTables[id][index] = _indexTable[value << (8 - bitDepth) & 0xff];
	}
	public void setWaveTableValue(int id, int index, int value, int bitDepth)
	{
		if (bitDepth > 8 || bitDepth < 1) bitDepth = 8;
		waveTables[id][index] = _indexTable[value << (8 - bitDepth) & 0xff];
	}
	public void setWaveTableValueN(int id, int index, float value){ setWaveTableValueN(id, index, value, 8); }
	public void setWaveTableValueN(int id, int index, float value, int bitDepth)
	{
		if (bitDepth > 8 || bitDepth < 1) bitDepth = 8;
		int v = (int)(value * 127.0f) + 128;
		waveTables[id][index] = _indexTable[v >> (8 - bitDepth)];
	}
	public SampleData getSampleData(int id){ return sampleDatas[id]; }
	public void setSampleData(int id, SampleData sampleData, boolean deepCopy)
	{
		allSoundStop();
		sampleDatas[id].copy(sampleData, deepCopy);
	}
	public void loadPCMSamples(int id, float[] wave, int ch)
	{
		allSoundStop();
		sampleDatas[id].loadPCMSamples(wave, ch);
	}
	public void loadPCMData(int id, byte[] dat, int position, int sampleFrames, int ch, int bitRate, boolean sign, boolean bigEndian)
	{
		allSoundStop();
		sampleDatas[id].loadPCMData(dat, position, sampleFrames, ch, bitRate, sign, bigEndian);
	}
	public void clearSampleData(int id)
	{
		allSoundStop();
		sampleDatas[id].clear();
	}
	@Override
	public void factoryDataReset()
	{
		if (_voiceUnitExt != null) allSoundStop();
		// waveData
		final double _PID2 = 3.141592653589793 * 2.0;
		int[] _simpleWave = new int[WAVET_CONT_SIZE];
		double _sizeD = (double)_waveTableSize;
		for (int i = 0; i < _waveTableSize; i++)
		{
			double _nv = Math.sin((i / _sizeD) * _PID2);
			int _iv = (int)(_nv * 127.5 - 0.5) + 128;
			_simpleWave[i] = _indexTable[_iv];
		}
		if (waveTables == null)waveTables = new int[128][];
		for (int i = 0; i < 128; i++)
		{
			if (waveTables[i] == null)waveTables[i] = new int[WAVET_CONT_SIZE];
			for (int j = 0; j < _waveTableSize; j++) waveTables[i][j] = _simpleWave[j];
		}
		// Preset
		if (tonePresets == null)tonePresets = new TonePresetWT[256];
		for (int i = 0; i < 256; i++)
		{
			if (tonePresets[i] == null)tonePresets[i] = new TonePresetWT();
			else tonePresets[i].reset();
		}
	// Sampler
		if (sampleDatas == null)sampleDatas = new SampleData[128];
		if (samplerPresets == null)samplerPresets = new SamplerPreset[128];
		for (int i = 0; i < 128; i++)
		{
			if (sampleDatas[i] == null)sampleDatas[i] = new SampleData();
			else sampleDatas[i].clear();
			if (samplerPresets[i] == null)samplerPresets[i] = new SamplerPreset();
			else samplerPresets[i].reset();
		}
	}
// Synthesizer - main -----
	private VoiceUnitWT[] _voiceUnitExt = null;
// Constructor
	public SynthesizerWT(){ this(32, 32, 440.0f, 1.0f, true); }
	public SynthesizerWT(int tblSize, int maxPoly, float refPitch, float masVol, boolean privateTableData)
	{
	// super ===
		super(maxPoly, refPitch, masVol);
	// SynthesizerWT ===
		_modelID = 0x20;
		_maxPMN = 3;
		int v, _b;
		for (v = 8, _b = 3; v < WAVET_CONT_SIZE; v <<= 1, _b++) if (v >= tblSize) break;
		_waveTableSizeBits = _b;
		_waveTableSize = v;
		// create
		_synthData = (privateTableData)? new SynthData() : SynthData.getInstance();
		_resetIndexTable();
		factoryDataReset();
		_recreateVU(_maxPolyphony);
		_resetKeyParams();
	}
// MidiEvent ---
	@Override
	protected void _noteOn(MidiEvent ev)
	{
		VoicePart _pt = voiceParts[ev.channel];
		final int _start = _nton;
		while ((_nton = _nton + 1 & _ntonM) != _start)
		{
			VoiceUnitWT _vu = _voiceUnitExt[_nton];
			if (_vu.state == 0)
			{
				_vu.startFrame = ev.offsetFrames;
				if (_pt.mode == 3)_vu.keyOnSampler(ev.channel, ev.value0, ev.value1);
				else _vu.keyOn(ev.channel, ev.value0, ev.value1);
				return;
			}
		}
			
		while ((_nton = _nton + 1 & _ntonM) != _start)
		{
			VoiceUnitWT _vu = _voiceUnitExt[_nton];
			// (KeyOff || SoundOff) -->
			if ((_vu.state & 0x60) != 0)
			{
				_vu.startFrame = ev.offsetFrames;
				if (_pt.mode == 3)_vu.keyOnSampler(ev.channel, ev.value0, ev.value1);
				else _vu.keyOn(ev.channel, ev.value0, ev.value1);
				return;
			}
		}
		
		while ((_nton = _nton + 1 & _ntonM) != _start)
		{
			VoiceUnitWT _vu = _voiceUnitExt[_nton];
			// KeyOn -->
			if ((_vu.state & 0x10) != 0)
			{
				_vu.startFrame = ev.offsetFrames;
				if (_pt.mode == 3)_vu.keyOnSampler(ev.channel, ev.value0, ev.value1);
				else _vu.keyOn(ev.channel, ev.value0, ev.value1);
				return;
			}
		}
	}
	@Override
	protected void _readSysExData(MidiEvent ev)
	{
		final byte[] _data = ev.data;
		final int _dataEnd = ev.dataSize - 1;  // minus F7
		if (_dataEnd < 8)return;
		int p = 0, v = 0;
		if ((_data[p++]&0xff) != _manufacID) return;
		if ((_data[p++]&0xff) != _deiceID) return;
		if ((_data[p++]&0xff) != _modelID) return;
		int b = _data[p++]&0xff;
		switch (b)	// Address 1
		{
			case 0x00:
				b = (_data[p++]&0xff)<<8 | (_data[p++]&0xff);
				switch (b)	// Address 2,3
				{
					case 0x0000 : setMasterVolume(_masterVolumeTable[(_data[p++]&127)]); if (p >= _dataEnd) break;
					case 0x0001 : setReferencePitch(440.0f + ((_data[p++]&127) - 64) * 0.5f); break;
				}
				break;
		// Part
			case 0x02:
				v = _data[p++] & 15;	// Address 2 [ CH ]
				b = _data[p++]&0x7f;
				switch (b)	// Address 3
				{
					case 0x00 : setDefaultPartMode(v, _data[p++]); if (p >= _dataEnd) break;
					case 0x01 : setPartMode(v, _data[p++]); break;
				}
				break;
		// Inst (Drum map)
			case 0x05: v = 128;
		// Inst (Program map)
			case 0x04:
				TonePresetWT _preset = tonePresets[v | (_data[p++]&127)];	// Address 2 [ ID ]
				b = (_data[p++]&0xff);
				switch (b)	// Address 3
				{
					case 0x00 : _preset.toneName[0]  = (char)_data[p++]; if (p >= _dataEnd) break;
					case 0x01 : _preset.toneName[1]  = (char)_data[p++]; if (p >= _dataEnd) break;
					case 0x02 : _preset.toneName[2]  = (char)_data[p++]; if (p >= _dataEnd) break;
					case 0x03 : _preset.toneName[3]  = (char)_data[p++]; if (p >= _dataEnd) break;
					case 0x04 : _preset.toneName[4]  = (char)_data[p++]; if (p >= _dataEnd) break;
					case 0x05 : _preset.toneName[5]  = (char)_data[p++]; if (p >= _dataEnd) break;
					case 0x06 : _preset.toneName[6]  = (char)_data[p++]; if (p >= _dataEnd) break;
					case 0x07 : _preset.toneName[7]  = (char)_data[p++]; if (p >= _dataEnd) break;
					case 0x08 : _preset.toneName[8]  = (char)_data[p++]; if (p >= _dataEnd) break;
					case 0x09 : _preset.toneName[9]  = (char)_data[p++]; if (p >= _dataEnd) break;
					case 0x0A : _preset.toneName[10] = (char)_data[p++]; if (p >= _dataEnd) break;
					case 0x0B : _preset.toneName[11] = (char)_data[p++]; if (p >= _dataEnd) break;
					case 0x0C : _preset.toneName[12] = (char)_data[p++]; if (p >= _dataEnd) break;
					case 0x0D : _preset.toneName[13] = (char)_data[p++]; if (p >= _dataEnd) break;
					case 0x0E : _preset.toneName[14] = (char)_data[p++]; if (p >= _dataEnd) break;
					case 0x0F : _preset.toneName[15] = (char)_data[p++]; if (p >= _dataEnd) break;

					case 0x10 : _preset.key     = _data[p++] & 127; if (p >= _dataEnd) break;
					case 0x11 : _preset.pan     = _data[p++] & 127; if (p >= _dataEnd) break;
					case 0x12 : _preset.lfoFreq = _preset.lfoFreq & 0x3f80 | (_data[p++]&127);      if (p >= _dataEnd) break;
					case 0x13 : _preset.lfoFreq = _preset.lfoFreq & 0x007f | (_data[p++]&127) << 7; if (p >= _dataEnd) break;
					case 0x14 :_preset.strc     = _data[p++] & 3;   if (p >= _dataEnd) break;
					case 0x15 : _preset.tune    = _preset.tune & 0x3f80 | (_data[p++]&127);      if (p >= _dataEnd) break;
					case 0x16 : _preset.tune    = _preset.tune & 0x007f | (_data[p++]&127) << 7; if (p >= _dataEnd) break;
					case 0x17 : _preset.lms     = _data[p++] & 15;  if (p >= _dataEnd) break;
					case 0x18 : _preset.wave    = _data[p++] & 127; if (p >= _dataEnd) break;
					case 0x19 : _preset.tl      = _data[p++] & 127; if (p >= _dataEnd) break;
					case 0x1A : _preset.ar      = _data[p++] & 127; if (p >= _dataEnd) break;
					case 0x1B : _preset.dr      = _data[p++] & 127; if (p >= _dataEnd) break;
					case 0x1C : _preset.sl      = _data[p++] & 127; if (p >= _dataEnd) break;
					case 0x1D : _preset.rr      = _data[p++] & 127; if (p >= _dataEnd) break;
					case 0x1E : _preset.kf      = _data[p++] & 3;   if (p >= _dataEnd) break;
					case 0x1F : _preset.kfStart = _data[p++] & 127; break;
				}
				break;
		// Percussion
			case 0x06:
				SamplerPreset _smplPrst = samplerPresets[(_data[p++]&127)];	// Address 2 [ ID ]
				b = (_data[p++]&0xff);
				switch (b)	// Address 3
				{
					case 0x00 : _smplPrst.toneName[0]  = (char)_data[p++]; if (p >= _dataEnd) break;
					case 0x01 : _smplPrst.toneName[1]  = (char)_data[p++]; if (p >= _dataEnd) break;
					case 0x02 : _smplPrst.toneName[2]  = (char)_data[p++]; if (p >= _dataEnd) break;
					case 0x03 : _smplPrst.toneName[3]  = (char)_data[p++]; if (p >= _dataEnd) break;
					case 0x04 : _smplPrst.toneName[4]  = (char)_data[p++]; if (p >= _dataEnd) break;
					case 0x05 : _smplPrst.toneName[5]  = (char)_data[p++]; if (p >= _dataEnd) break;
					case 0x06 : _smplPrst.toneName[6]  = (char)_data[p++]; if (p >= _dataEnd) break;
					case 0x07 : _smplPrst.toneName[7]  = (char)_data[p++]; if (p >= _dataEnd) break;
					case 0x08 : _smplPrst.toneName[8]  = (char)_data[p++]; if (p >= _dataEnd) break;
					case 0x09 : _smplPrst.toneName[9]  = (char)_data[p++]; if (p >= _dataEnd) break;
					case 0x0A : _smplPrst.toneName[10] = (char)_data[p++]; if (p >= _dataEnd) break;
					case 0x0B : _smplPrst.toneName[11] = (char)_data[p++]; if (p >= _dataEnd) break;
					case 0x0C : _smplPrst.toneName[12] = (char)_data[p++]; if (p >= _dataEnd) break;
					case 0x0D : _smplPrst.toneName[13] = (char)_data[p++]; if (p >= _dataEnd) break;
					case 0x0E : _smplPrst.toneName[14] = (char)_data[p++]; if (p >= _dataEnd) break;
					case 0x0F : _smplPrst.toneName[15] = (char)_data[p++]; if (p >= _dataEnd) break;
				
					case 0x10 : _smplPrst.smplNum = _data[p++] & 127; if (p >= _dataEnd) break;
					case 0x11 : _smplPrst.pan     = _data[p++] & 127; if (p >= _dataEnd) break;
					case 0x12 : _smplPrst.octv    = _data[p++] & 15;  if (p >= _dataEnd) break;
					case 0x13 : _smplPrst.tune    = _smplPrst.tune & 0x3f80 | (_data[p++]&127);      if (p >= _dataEnd) break;
					case 0x14 : _smplPrst.tune    = _smplPrst.tune & 0x007f | (_data[p++]&127) << 7; if (p >= _dataEnd) break;
					case 0x15 : _smplPrst.tl      = _data[p++] & 127; if (p >= _dataEnd) break;
					case 0x16 : _smplPrst.ar      = _data[p++] & 127; if (p >= _dataEnd) break;
					case 0x17 : _smplPrst.dr      = _data[p++] & 127; if (p >= _dataEnd) break;
					case 0x18 : _smplPrst.sl      = _data[p++] & 127; if (p >= _dataEnd) break;
					case 0x19 : _smplPrst.rr      = _data[p++] & 127; break;
				}
				break;
		// Data
			case 0x08:
				v =  _data[p++] & 127;	// Address 2 [ ID ]
				for (int i = _data[p++] & 127; i < WAVET_CONT_SIZE; i++)
				{
					waveTables[v][i] = _indexTable[(_data[p++] & 127) << 1];
					if (p >= _dataEnd) break;
				}
				break;
		}
	}
}